# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from functools import lru_cache
from pathlib import Path
from typing import Optional

from bluecellulab.circuit.config.sections import Conditions, ConnectionOverrides
from bluecellulab.stimulus.circuit_stimulus_definitions import Stimulus

from bluepysnap import Simulation as SnapSimulation


class SonataSimulationConfig:
    """Sonata implementation of SimulationConfig protocol."""
    _connection_overrides: list[ConnectionOverrides] = []

    def __init__(self, config: str | Path | SnapSimulation) -> None:
        if isinstance(config, (str, Path)):
            if not Path(config).exists():
                raise FileNotFoundError(f"Config file {config} not found.")
            else:
                self.impl = SnapSimulation(config)
        elif isinstance(config, SnapSimulation):
            self.impl = config
        else:
            raise TypeError("Invalid config type.")

    def get_all_projection_names(self) -> list[str]:
        unique_names = {
            n
            for n in self.impl.circuit.nodes
            if self.impl.circuit.nodes[n].type == "virtual"
        }
        return list(unique_names)

    def get_all_stimuli_entries(self) -> list[Stimulus]:
        result: list[Stimulus] = []
        inputs = self.impl.config.get("inputs")
        if inputs is None:
            return result
        for value in inputs.values():
            stimulus = Stimulus.from_sonata(value)
            if stimulus:
                result.append(stimulus)
        return result

    @lru_cache(maxsize=1)
    def condition_parameters(self) -> Conditions:
        """Returns parameters of global condition block of sonataconfig."""
        condition_object = self.impl.conditions
        return Conditions.from_sonata(condition_object)

    @lru_cache(maxsize=1)
    def _connection_entries(self) -> list[ConnectionOverrides]:
        result: list[ConnectionOverrides] = []
        if "connection_overrides" not in self.impl.config:
            return result
        conn_overrides: list = self.impl.config["connection_overrides"]
        if conn_overrides is None:
            return result
        for conn_entry in conn_overrides:
            result.append(ConnectionOverrides.from_sonata(conn_entry))
        return result

    def connection_entries(self) -> list[ConnectionOverrides]:
        return self._connection_entries() + self._connection_overrides

    @property
    def base_seed(self) -> int:
        return self.impl.run.random_seed

    @property
    def synapse_seed(self) -> int:
        return self.impl.run.synapse_seed

    @property
    def ionchannel_seed(self) -> int:
        return self.impl.run.ionchannel_seed

    @property
    def stimulus_seed(self) -> int:
        return self.impl.run.stimulus_seed

    @property
    def minis_seed(self) -> int:
        return self.impl.run.minis_seed

    @property
    def rng_mode(self) -> str:
        """Only Random123 is supported in SONATA."""
        return "Random123"

    @property
    def spike_threshold(self) -> float:
        return self.impl.run.spike_threshold

    @property
    def spike_location(self) -> str:
        return self.impl.conditions.spike_location.name

    @property
    def duration(self) -> Optional[float]:
        return self.impl.run.tstop

    @property
    def dt(self) -> float:
        return self.impl.run.dt

    @property
    def forward_skip(self) -> Optional[float]:
        """forward_skip is removed from SONATA."""
        return None

    @property
    def celsius(self) -> float:
        value = self.condition_parameters().celsius
        return value if value is not None else 34.0

    @property
    def v_init(self) -> float:
        value = self.condition_parameters().v_init
        return value if value is not None else -65.0

    @property
    def output_root_path(self) -> str:
        return self.impl.config["output"]["output_dir"]

    @property
    def extracellular_calcium(self) -> Optional[float]:
        return self.condition_parameters().extracellular_calcium

    def add_connection_override(
        self,
        connection_override: ConnectionOverrides
    ) -> None:
        self._connection_overrides.append(connection_override)



"""Module responsible of circuit access."""

from .synapse_properties import SynapseProperty
from .node_id import CellId
from .circuit_access import BluepyCircuitAccess, CircuitAccess, SonataCircuitAccess, EmodelProperties
from .simulation_access import BluepySimulationAccess, SimulationAccess, SonataSimulationAccess
from .validate import SimulationValidator



from __future__ import annotations
from abc import ABC, abstractmethod
import matplotlib.pyplot as plt
import numpy as np


class Stimulus(ABC):
    def __init__(self, dt: float) -> None:
        self.dt = dt

    @property
    @abstractmethod
    def time(self) -> np.ndarray:
        """Time values of the stimulus."""
        ...

    @property
    @abstractmethod
    def current(self) -> np.ndarray:
        """Current values of the stimulus."""
        ...

    def __len__(self) -> int:
        return len(self.time)

    @property
    def stimulus_time(self) -> float:
        return len(self) * self.dt

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(dt={self.dt})"

    def plot(self, ax=None, **kwargs):
        if ax is None:
            ax = plt.gca()
        ax.plot(self.time, self.current, **kwargs)
        ax.set_xlabel("Time (ms)")
        ax.set_ylabel("Current (nA)")
        ax.set_title(self.__class__.__name__)
        return ax

    def __add__(self, other: Stimulus) -> CombinedStimulus:
        """Override + operator to concatenate Stimulus objects."""
        if self.dt != other.dt:
            raise ValueError(
                "Stimulus objects must have the same dt to be concatenated"
            )
        if len(self.time) == 0:
            return CombinedStimulus(other.dt, other.time, other.current)
        elif len(other.time) == 0:
            return CombinedStimulus(self.dt, self.time, self.current)
        else:
            # shift other time
            other_time = other.time + self.time[-1] + self.dt
            combined_time = np.concatenate([self.time, other_time])
            # Concatenate the current arrays
            combined_current = np.concatenate([self.current, other.current])
            return CombinedStimulus(self.dt, combined_time, combined_current)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Stimulus):
            return NotImplemented
        else:
            return (
                np.allclose(self.time, other.time)
                and np.allclose(self.current, other.current)
                and self.dt == other.dt
            )


class CombinedStimulus(Stimulus):
    """Represents the Stimulus created by combining multiple stimuli."""

    def __init__(self, dt: float, time: np.ndarray, current: np.ndarray) -> None:
        super().__init__(dt)
        self._time = time
        self._current = current

    @property
    def time(self) -> np.ndarray:
        return self._time

    @property
    def current(self) -> np.ndarray:
        return self._current


class Empty(Stimulus):
    """Represents empty stimulus (all zeros) that has no impact on the cell.

    This is required by some Stimuli that expect the cell to rest.
    """

    def __init__(self, dt: float, duration: float) -> None:
        super().__init__(dt)
        self.duration = duration

    @property
    def time(self) -> np.ndarray:
        return np.arange(0.0, self.duration, self.dt)

    @property
    def current(self) -> np.ndarray:
        return np.zeros_like(self.time)


class Flat(Stimulus):
    def __init__(self, dt: float, duration: float, amplitude: float) -> None:
        super().__init__(dt)
        self.duration = duration
        self.amplitude = amplitude

    @property
    def time(self) -> np.ndarray:
        return np.arange(0.0, self.duration, self.dt)

    @property
    def current(self) -> np.ndarray:
        return np.full_like(self.time, self.amplitude)


class Slope(Stimulus):
    def __init__(
        self, dt: float, duration: float, amplitude_start: float, amplitude_end: float
    ) -> None:
        super().__init__(dt)
        self.duration = duration
        self.amplitude_start = amplitude_start
        self.amplitude_end = amplitude_end

    @property
    def time(self) -> np.ndarray:
        return np.arange(0.0, self.duration, self.dt)

    @property
    def current(self) -> np.ndarray:
        return np.linspace(self.amplitude_start, self.amplitude_end, len(self.time))


class Step(Stimulus):

    def __init__(self):
        raise NotImplementedError(
            "This class cannot be instantiated directly. "
            "Please use the class methods 'amplitude_based' "
            "or 'threshold_based' to create objects."
        )

    @classmethod
    def amplitude_based(
        cls,
        dt: float,
        pre_delay: float,
        duration: float,
        post_delay: float,
        amplitude: float,
    ) -> CombinedStimulus:
        """Create a Step stimulus from given time events and amplitude.

        Args:
            dt: The time step of the stimulus.
            pre_delay: The delay before the start of the step.
            duration: The duration of the step.
            post_delay: The time to wait after the end of the step.
            amplitude: The amplitude of the step.
        """
        return (
            Empty(dt, duration=pre_delay)
            + Flat(dt, duration=duration, amplitude=amplitude)
            + Empty(dt, duration=post_delay)
        )

    @classmethod
    def threshold_based(
        cls,
        dt: float,
        pre_delay: float,
        duration: float,
        post_delay: float,
        threshold_current: float,
        threshold_percentage: float,
    ) -> CombinedStimulus:
        """Creates a Step stimulus with respect to the threshold current.

        Args:

            dt: The time step of the stimulus.
            pre_delay: The delay before the start of the step.
            duration: The duration of the step.
            post_delay: The time to wait after the end of the step.
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
        """
        amplitude = threshold_current * threshold_percentage / 100
        res = cls.amplitude_based(
            dt,
            pre_delay=pre_delay,
            duration=duration,
            post_delay=post_delay,
            amplitude=amplitude,
        )
        return res


class Ramp(Stimulus):

    def __init__(self):
        raise NotImplementedError(
            "This class cannot be instantiated directly. "
            "Please use the class methods 'amplitude_based' "
            "or 'threshold_based' to create objects."
        )

    @classmethod
    def amplitude_based(
        cls,
        dt: float,
        pre_delay: float,
        duration: float,
        post_delay: float,
        amplitude: float,
    ) -> CombinedStimulus:
        """Create a Ramp stimulus from given time events and amplitudes.

        Args:
            dt: The time step of the stimulus.
            pre_delay: The delay before the start of the ramp.
            duration: The duration of the ramp.
            post_delay: The time to wait after the end of the ramp.
            amplitude: The final amplitude of the ramp.
        """
        return (
            Empty(dt, duration=pre_delay)
            + Slope(
                dt,
                duration=duration,
                amplitude_start=0.0,
                amplitude_end=amplitude,
            )
            + Empty(dt, duration=post_delay)
        )

    @classmethod
    def threshold_based(
        cls,
        dt: float,
        pre_delay: float,
        duration: float,
        post_delay: float,
        threshold_current: float,
        threshold_percentage: float,
    ) -> CombinedStimulus:
        """Creates a Ramp stimulus with respect to the threshold current.

        Args:

            dt: The time step of the stimulus.
            pre_delay: The delay before the start of the ramp.
            duration: The duration of the ramp.
            post_delay: The time to wait after the end of the ramp.
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
        """
        amplitude = threshold_current * threshold_percentage / 100
        res = cls.amplitude_based(
            dt,
            pre_delay=pre_delay,
            duration=duration,
            post_delay=post_delay,
            amplitude=amplitude,
        )
        return res


class StimulusFactory:
    def __init__(self, dt: float):
        self.dt = dt

    def step(
        self, pre_delay: float, duration: float, post_delay: float, amplitude: float
    ) -> Stimulus:
        return Step.amplitude_based(
            self.dt,
            pre_delay=pre_delay,
            duration=duration,
            post_delay=post_delay,
            amplitude=amplitude,
        )

    def ramp(
        self, pre_delay: float, duration: float, post_delay: float, amplitude: float
    ) -> Stimulus:
        return Ramp.amplitude_based(
            self.dt,
            pre_delay=pre_delay,
            duration=duration,
            post_delay=post_delay,
            amplitude=amplitude,
        )

    def ap_waveform(
        self, threshold_current: float, threshold_percentage: float = 220.0
    ) -> Stimulus:
        """Returns the APWaveform Stimulus object, a type of Step stimulus.

        Args:
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
        """
        pre_delay = 250.0
        duration = 50.0
        post_delay = 250.0
        return Step.threshold_based(
            self.dt,
            pre_delay=pre_delay,
            duration=duration,
            post_delay=post_delay,
            threshold_current=threshold_current,
            threshold_percentage=threshold_percentage,
        )

    def idrest(
        self,
        threshold_current: float,
        threshold_percentage: float = 200.0,
    ) -> Stimulus:
        """Returns the IDRest Stimulus object, a type of Step stimulus.

        Args:
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
        """
        pre_delay = 250.0
        duration = 1350.0
        post_delay = 250.0
        return Step.threshold_based(
            self.dt,
            pre_delay=pre_delay,
            duration=duration,
            post_delay=post_delay,
            threshold_current=threshold_current,
            threshold_percentage=threshold_percentage,
        )

    def iv(
        self,
        threshold_current: float,
        threshold_percentage: float = -40.0,
    ) -> Stimulus:
        """Returns the IV Stimulus object, a type of Step stimulus.

        Args:
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
        """
        pre_delay = 250.0
        duration = 3000.0
        post_delay = 250.0
        return Step.threshold_based(
            self.dt,
            pre_delay=pre_delay,
            duration=duration,
            post_delay=post_delay,
            threshold_current=threshold_current,
            threshold_percentage=threshold_percentage,
        )

    def fire_pattern(
        self,
        threshold_current: float,
        threshold_percentage: float = 200.0,
    ) -> Stimulus:
        """Returns the FirePattern Stimulus object, a type of Step stimulus.

        Args:
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
        """
        pre_delay = 250.0
        duration = 3600.0
        post_delay = 250.0
        return Step.threshold_based(
            self.dt,
            pre_delay=pre_delay,
            duration=duration,
            post_delay=post_delay,
            threshold_current=threshold_current,
            threshold_percentage=threshold_percentage,
        )

    def pos_cheops(
        self,
        threshold_current: float,
        threshold_percentage: float = 300.0,
    ) -> Stimulus:
        """A combination of pyramid shaped Ramp stimuli with a positive
        amplitude.

        Args:
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
        """
        delay = 250.0
        ramp1_duration = 4000.0
        ramp2_duration = 2000.0
        ramp3_duration = 1333.0
        inter_delay = 2000.0
        post_delay = 250.0

        amplitude = threshold_current * threshold_percentage / 100
        result = (
            Empty(self.dt, duration=delay)
            + Slope(self.dt, duration=ramp1_duration, amplitude_start=0.0, amplitude_end=amplitude)
            + Slope(self.dt, duration=ramp1_duration, amplitude_start=amplitude, amplitude_end=0.0)
            + Empty(self.dt, duration=inter_delay)
            + Slope(self.dt, duration=ramp2_duration, amplitude_start=0.0, amplitude_end=amplitude)
            + Slope(self.dt, duration=ramp2_duration, amplitude_start=amplitude, amplitude_end=0.0)
            + Empty(self.dt, duration=inter_delay)
            + Slope(self.dt, duration=ramp3_duration, amplitude_start=0.0, amplitude_end=amplitude)
            + Slope(self.dt, duration=ramp3_duration, amplitude_start=amplitude, amplitude_end=0.0)
            + Empty(self.dt, duration=post_delay)
        )
        return result

    def neg_cheops(
        self,
        threshold_current: float,
        threshold_percentage: float = 300.0,
    ) -> Stimulus:
        """A combination of pyramid shaped Ramp stimuli with a negative
        amplitude.

        Args:
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
        """
        delay = 1750.0
        ramp1_duration = 3333.0
        ramp2_duration = 1666.0
        ramp3_duration = 1111.0
        inter_delay = 2000.0
        post_delay = 250.0

        amplitude = - threshold_current * threshold_percentage / 100
        result = (
            Empty(self.dt, duration=delay)
            + Slope(self.dt, duration=ramp1_duration, amplitude_start=0.0, amplitude_end=amplitude)
            + Slope(self.dt, duration=ramp1_duration, amplitude_start=amplitude, amplitude_end=0.0)
            + Empty(self.dt, duration=inter_delay)
            + Slope(self.dt, duration=ramp2_duration, amplitude_start=0.0, amplitude_end=amplitude)
            + Slope(self.dt, duration=ramp2_duration, amplitude_start=amplitude, amplitude_end=0.0)
            + Empty(self.dt, duration=inter_delay)
            + Slope(self.dt, duration=ramp3_duration, amplitude_start=0.0, amplitude_end=amplitude)
            + Slope(self.dt, duration=ramp3_duration, amplitude_start=amplitude, amplitude_end=0.0)
            + Empty(self.dt, duration=post_delay)
        )
        return result



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Identifier for the circuit nodes.

Account for multi-population and the single.
"""

from __future__ import annotations
from typing import NamedTuple


CellId = NamedTuple("CellId", [("population_name", str), ("id", int)])


def create_cell_id(cell_id: int | tuple[str, int] | CellId) -> CellId:
    """Make a CellId from a tuple or an int."""
    if isinstance(cell_id, tuple):
        return CellId(*cell_id)
    else:
        return CellId("", cell_id)


def create_cell_ids(cell_ids: list[int] | list[tuple[str, int] | CellId]) -> list[CellId]:
    """Make a list of CellId from a list of tuple or int."""
    return [create_cell_id(cell_id) for cell_id in cell_ids]



"""A simple ball and stick model."""

import importlib_resources as resources
from bluecellulab.cell import Cell


def create_ball_stick() -> Cell:
    """Creates a ball and stick model by using package's hoc and asc."""
    hoc = resources.files("bluecellulab") / "cell/ballstick/emodel.hoc"
    morphology = resources.files("bluecellulab") / "cell/ballstick/morphology.asc"
    return Cell(template_path=hoc, morphology_path=morphology)



"""Neuron recordings related functions."""
from __future__ import annotations
from bluecellulab.type_aliases import NeuronSection


def section_to_voltage_recording_str(section: NeuronSection, segment=0.5) -> str:
    """Converts a section and segment to voltage recording string."""
    return f"neuron.h.{section.name()}({segment})._ref_v"



"""Package for performing single cell or network simulations and
experiments."""

try:
    import bluepy
    BLUEPY_AVAILABLE = True
except ImportError:
    BLUEPY_AVAILABLE = False

from bluecellulab.importer import import_hoc
from .verbosity import *
from .cell import Cell, create_ball_stick  # NOQA
from .circuit import EmodelProperties
from .connection import Connection  # NOQA
from .plotwindow import PlotWindow  # NOQA
from .dendrogram import Dendrogram  # NOQA
from .psection import PSection  # NOQA
from .psegment import PSegment  # NOQA
from .simulation import Simulation  # NOQA
from .rngsettings import RNGSettings  # NOQA
from .circuit_simulation import CircuitSimulation, CircuitSimulation  # NOQA
import neuron

from .simulation.neuron_globals import NeuronGlobals

logger.debug("Loading the hoc files.")
import_hoc(neuron)
_ = NeuronGlobals.get_instance()  # initiate the singleton



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module to access the simulation results."""

from __future__ import annotations
import logging
from pathlib import Path
from typing import Optional, Protocol

import h5py
import pandas as pd

from bluecellulab.circuit.config import SimulationConfig, SonataSimulationConfig
from bluecellulab.exceptions import ExtraDependencyMissingError

from bluecellulab import BLUEPY_AVAILABLE
if BLUEPY_AVAILABLE:
    import bluepy
    from bluecellulab.circuit.iotools import parse_outdat

from bluepysnap import Simulation as SnapSimulation
import numpy as np

from bluecellulab.circuit import CellId
from bluecellulab.circuit.config import BluepySimulationConfig


logger = logging.getLogger(__name__)


def _sample_array(arr: np.ndarray, ratio: float) -> np.ndarray:
    """Sample an array at a given time step.

    Args:
        arr: Array to sample.
        ratio: The ratio between the time step used for sampling and the time step used in the simulation
                (the time step should be a multiple of the simulation time step)

    Returns:
        Array sampled at the given time step.
    """
    if ratio == 1:
        return arr
    elif not np.isclose(ratio, round(ratio)):
        raise ValueError("The ratio is not close to a whole number. "
                         "The time step should be a multiple of the simulation time step.")
    return arr[::round(ratio)]


class SimulationAccess(Protocol):
    """Protocol that defines the simulation access layer."""

    impl: bluepy.Simulation | SnapSimulation

    def get_soma_voltage(
        self, cell_id: CellId, t_start: float, t_end: float, t_step: Optional[float] = None
    ) -> np.ndarray:
        raise NotImplementedError

    def get_soma_time_trace(self, t_step: Optional[float] = None) -> np.ndarray:
        raise NotImplementedError

    def get_spikes(self) -> dict[CellId, np.ndarray]:
        """Get spikes from the main simulation."""
        raise NotImplementedError


class BluepySimulationAccess:
    """Bluepy implementation of SimulationAccess protocol."""

    def __init__(self, sim_config: str | Path | SimulationConfig) -> None:
        """Initialize the simulation access object."""
        if not BLUEPY_AVAILABLE:
            raise ExtraDependencyMissingError("bluepy")
        if isinstance(sim_config, BluepySimulationConfig):
            sim_config = sim_config.impl
        elif isinstance(sim_config, Path):
            sim_config = str(sim_config)
        if isinstance(sim_config, str) and not Path(sim_config).exists():
            raise FileNotFoundError(
                f"Circuit config file {sim_config} not found.")

        self.impl = bluepy.Simulation(sim_config)
        self._config = BluepySimulationConfig(sim_config)  # type: ignore

    def get_soma_voltage(
        self, cell_id: CellId, t_start: float, t_end: float, t_step: Optional[float] = None
    ) -> np.ndarray:
        """Retrieve the soma voltage of main simulation."""
        gid = cell_id.id
        arr = (
            self.impl.report("soma")
            .get_gid(gid, t_start=t_start, t_end=t_end)
            .to_numpy()
        )
        if t_step is not None:
            ratio = t_step / self._config._soma_report_dt
            arr = _sample_array(arr, ratio)
        return arr

    def get_soma_time_trace(self, t_step: Optional[float] = None) -> np.ndarray:
        """Retrieve the time trace from the main simulation."""
        report = self.impl.report('soma')
        arr = report.get_gid(report.gids[0]).index.to_numpy()
        if t_step is not None:
            ratio = t_step / self._config._soma_report_dt
            arr = _sample_array(arr, ratio)
        return arr

    def get_spikes(self) -> dict[CellId, np.ndarray]:
        outdat_path = Path(self._config.output_root_path) / "out.dat"
        return parse_outdat(outdat_path)


class SonataSimulationAccess:
    """Sonata implementation of SimulationAccess protocol."""

    def __init__(self, sim_config: str | Path | SimulationConfig) -> None:
        """Initialize SonataCircuitAccess object."""
        if isinstance(sim_config, (str, Path)) and not Path(sim_config).exists():
            raise FileNotFoundError(f"Circuit config file {sim_config} not found.")

        if isinstance(sim_config, SonataSimulationConfig):
            self.impl = sim_config.impl
        else:
            self.impl = SnapSimulation(sim_config)

    def get_soma_voltage(
        self, cell_id: CellId, t_start: float, t_end: float, t_step: Optional[float] = None
    ) -> np.ndarray:
        report = self.impl.reports["soma"].filter(cell_id.id, t_start, t_end)
        arr = report.report[cell_id.population_name][cell_id.id].values
        if t_step is not None:
            ratio = t_step / self.impl.dt
            arr = _sample_array(arr, ratio)
        return arr

    def get_soma_time_trace(self, t_step: Optional[float] = None) -> np.ndarray:
        report = self.impl.reports["soma"]
        arr = report.filter().report.index.values
        if t_step is not None:
            ratio = t_step / self.impl.dt
            arr = _sample_array(arr, ratio)
        return arr

    def get_spikes(self) -> dict[CellId, np.ndarray]:
        spike_report = self.impl.spikes
        filtered_report = spike_report.filter()
        # filtered_report.head(2)
        #         ids population
        # times
        # 10.1     0  hippocampus_neurons
        # 10.1     1  hippocampus_neurons

        # convert it to dict, where key is CellId made of (population and ids columns)
        # and value is np.array of spike times
        outdat = filtered_report.report.groupby(["population", "ids"])
        outdat = outdat.apply(lambda x: x.index.values)
        outdat.index = [CellId(pop_name, idx) for (pop_name, idx) in outdat.index]
        return outdat.to_dict()


def get_synapse_replay_spikes(f_name: str) -> dict:
    """Read the .dat file containing the spike replays."""
    with h5py.File(f_name, 'r') as f:
        # Access the timestamps and node_ids datasets
        timestamps = f['/spikes/All/timestamps'][:]
        node_ids = f['/spikes/All/node_ids'][:]

        spikes = pd.DataFrame(data={'t': timestamps, 'node_id': node_ids})
        spikes = spikes.astype({"node_id": int})

    if (spikes["t"] < 0).any():
        logger.warning("Found negative spike times... Clipping them to 0")
        spikes["t"].clip(lower=0., inplace=True)
    return spikes.groupby("node_id")["t"].apply(np.array).to_dict()



"""Simulation configuration file parser module."""

from .definition import SimulationConfig
from .bluepy_simulation_config import BluepySimulationConfig
from .sonata_simulation_config import SonataSimulationConfig
